#!/usr/bin/env python
"""
Scummbler v2

    This file is part of Scummbler.

    Scummbler is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Scummbler is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Scummbler.  If not, see <http://www.gnu.org/licenses/>.

Laurence Dougal Myers
www.jestarjokin.net

December 2008 - May 2009

Parses & compiles SCUMM V5 scripts.

Gotchas:
 - Labels only support alphanumeric characters and the underscore ("_") character.
 - "Local", "Var", and "Bit" are case sensitive and must begin with a capital letter.
 - The last item in a script should always be "stopObjectCode()" - maybe I could automate this?
 - "If" statements with no instructions after them will compile, but will create a runtime error 
   (because "ifs" are actually jumps, and the way I generate them it will try to jump to a
   non-existant only situation this could occur is if the above "gotcha" is violated.
 - You probably should not have a line containing just a label and a comment. It seems to
   work from testing, but I don't fully comprehend the grammar I wrote, so the world MAY end.
 - startScript is tricky. You can have opcode, or extra parameters R and F, but if you have
   the input opcode present, it will be ORed with the output opcode (which may already have
   flags from the extra "R" and "F" parameters). So, if you're using R and F, don't use the
   descumm opcodes.
   
Potential Tasks:
 - Compiler niceties, like arithmetic on constants, selecting "delay" or "delayVariable"
   based on the argument it's given...
 - Syntactic sugar, like joining multiple "if" statements with "and", "or".
 - Better parse error handling & messages (may not be possible with PyParsing); this is
   really VERY needed, as I've found with test scripts that if there's a problem within
   an "if" block, the error currently points to the start of the block, not the problem.
 
Need real example for comparison:
 - isActorInBox
 - most of roomOps
"""
import os.path
import sys
from pyparsing import *
from scummbler_opcodes import *
from scummbler_misc import *
from scummbler_lexicon import *
import scummbler_autogram

# Some globals for resolving jumps
# Jumps will be resolved like so:
#  - When we find a jump, note the destination label and the current line number
#     (there may be multiple jumps to the same label, so store a list of jumps
#       for each label)
#  - When we find a label, note the label name and the current line number
#  - After parsing, we resolve each jump, first checking each destination label.
#  - If the label does not exist in labelTable, we didn't see it in parsing
#     (error in the script).
#  - For each jump of that label, we get the line number containing the label,
#     and resolve the difference in bytes between the lines.
jumpTable = {} # key = label, item = line num containing jumping instruction
labelTable = {} # key = label, item = line num containing the label
ifStack = [] # used for resolving nested if statements
loopStack = [] # used for resolving loops (TODO: replace ifStack)
currIfBlock = None
scriptType = "SCRP"

# Not sure about these limits, but you can't jump more than a word.
maxJump = 0x7FFF
minJump = -0x7FFF

#__________________
# Start Grammar {

# Inline opcodes (not function calls)
grInlineOperation = Group(leVariable)("target") + oneOf("+= -= /= *= &= |= =")("operation") + Group(leVarOrWord)("value")
grIncDec = Group(leVariable)("target") + oneOf("++ --")("operation") # not sure if Bits can be incremented/decremented

# Jumps
grJumpLabel = SupLit("[") + leLabel("label") + SupLit("]")
grJumpGoto = SupLit("goto") + leLabel("target")
grJumpExpression = grJumpGoto

# Comparison & conditional jump/"if" related opcodes
grIfComparison = Group(leVariable)("arg1") + oneOf(" ".join(opComparisons.keys()))("comparator") + Group(leVarOrWord)("arg2")
grIfZeroEquality = Optional(Literal("!")("!")) + leVariable
grIfClassOfIs = Literal("classOfIs")("function") + SupLit("(") + Group(leVarOrWord)("testobject") + SupLit(",") + leArgList("classes") ^ Optional(SupLit(")")) # descumm forgets to add the last parenthesis
grIfActorInBox = Literal("isActorInBox")("function") + SupLit("(") + Group(leVarOrByte)("actor") + SupLit(",") + Group(leVarOrByte)("box") + SupLit(")")
grIfState = Literal("getState")("function") + SupLit("(") + Group(leVarOrWord)("testobject") + SupLit(")") + (Literal("==") | Literal("!="))("comparator") + Group(leVarOrByte)("state") + SupLit(")")
grBooleanExpression = grIfComparison ^ grIfZeroEquality ^ grIfClassOfIs ^ grIfActorInBox ^ grIfState
grJumpUnless = SupLit("unless") + SupLit("(") + (grBooleanExpression)("test") + SupLit(")") + SupLit("goto") + leLabel("target")
grIfStart = SupLit("if") + SupLit("(") + (grBooleanExpression) + SupLit(")") + SupLit("{")
grIfElse = SupLit("}") + SupLit("else")
grIfEnd = SupLit("}")

# Other control flow
grWhileStart = SupLit("while") + SupLit("(") + grBooleanExpression + SupLit(")") + SupLit("{")
grWhileEnd = SupLit("}") # obviously this conflicts with grIfEnd...
grDoWhileStart = SupLit("do") + SupLit("{")
grDoWhileEnd = SupLit("}") + SupLit("while") + SupLit("(") + grBooleanExpression + SupLit(")")
grForLoopStart = (SupLit("for") + SupLit("(") +
                  Optional(grInlineOperation)("init") + SupLit(";") + 
                  Group(grBooleanExpression)("test") + SupLit(";") + 
                  Optional(grInlineOperation | grIncDec)("increm") +
                  SupLit(")") + SupLit("{")) # "For" loops only allow basic inline maths for now
grForLoopEnd = SupLit("}")


# This used to be below but I need it for startScript
grDescummOpcode = SupLit("(") + (leHexByte | leLabel | Literal("**"))("dop") + SupLit(")")

# Functions that can't be automatically generated due to quirks in their grammar.

# delay has its own type of argument, a 24-bit constant
grFunc_delay = Literal("delay")("function") + SupLit("(") + Group(leDelay)("arg1") + SupLit(")")
# doSentence normally takes 3 params, except when it only takes 1 (which is 0xFE, descumm outputs it as "STOP")
grFunc_doSentence = (Literal("doSentence")("function") + SupLit("(") +
                     (Literal("STOP")("arg1") | (Group(leVarOrByte)("arg1") + SupLit(",") +
                                                 Group(leVarOrWord)("arg2") + SupLit(",") +
                                                 Group(leVarOrWord)("arg3")
                                                 )
                     ) + SupLit(")"))
# drawBox has aux opcodes
grFunc_drawBox = ( Literal("drawBox")("function") + SupLit("(") +
                  Group(leVarOrWord)("left") + SupLit(",") + 
                  Group(leVarOrWord)("top") + SupLit(",") + 
                  Group(leVarOrWord)("right") + SupLit(",") +
                  Group(leVarOrWord)("bottom") + SupLit(",") +
                  Group(leVarOrByte)("colour") + SupLit(")") )
# drawObject omits opening bracket of 0x1F (unnamed sub-opcode) but accidentally leaves in the closing bracket.
# Unfortunately, I have to define the whole of drawObject here (although it's based on auto-generated code).
grFunc_drawObject_setXY = Literal("setXY")("function") + SupLit("(") + Group(leVarOrWord)("arg1") + SupLit(",") + Group(leVarOrWord)("arg2") + SupLit(")")
grFunc_drawObject_setImage = Literal("setImage")("function") + SupLit("(") + Group(leVarOrWord)("arg1") + SupLit(")")
grFunc_drawObject_SO = grFunc_drawObject_setXY | grFunc_drawObject_setImage
# Make the whole of "arg2" to support the strange "empty" function - if it's missing, parse action knows what to do
grFunc_drawObject = (Literal("drawObject")("function") + SupLit("(") +
                     Group(leVarOrWord)("arg1") +
                     Optional( Suppress(Optional(Literal(","))) + grFunc_drawObject_SO("arg2") ) +
                     SupLit(")") + Optional(SupLit(")")))
# loadRoomWithEgo has two params and two constants
grFunc_loadRoomWithEgo = (Literal("loadRoomWithEgo")("function") + SupLit("(") +
                          Group(leVarOrWord)("object") + SupLit(",") +
                          Group(leVarOrByte)("room") + SupLit(",") +
                          leWord("x") + SupLit(",") +
                          leWord("y") +
                          SupLit(")"))
# oldRoomEffect is dodgy
grFunc_oldRoomEffect = ((Literal("oldRoomEffect-set") | Literal("oldRoomEffect-fadein"))("function") + SupLit("(") +
                        Group(leVarOrWord)("effect") +
                        SupLit(")")
                        )
# override is like a soft bunny without ears (no parentheses) (except those) (and those)
grFunc_override = (Literal("beginOverride") | Literal("endOverride"))("function") + Optional(SupLit("(") + SupLit(")"))
# pickupObject can be one of two opcodes, one which takes a word and a byte, the other ("pickupObjectOld") only takes a word.
grFunc_pickupObject = Literal("pickupObject")("function") + SupLit("(") + Group(leVarOrWord)("arg1") + Optional( SupLit(",") + Group(leVarOrByte)("arg2") ) + SupLit(")")
# "printEgo" re-uses the sub-opcodes for "print"
grFunc_printEgo = Literal("printEgo")("function") + SupLit("(") + SupLit("[") + Optional(delimitedList(scummbler_autogram.grFunc_print_SO))("arg1") + SupLit("]")  + SupLit(")")
# pseudoRoom takes one constant followed by a variable number of constants.
# Each res item is ORed with 0x80. A value of 0x80 is output as "IG" in descumm.
grFunc_PseudoRoom = (Literal("PseudoRoom")("function") + SupLit("(") +
                     leByte("val") + SupLit(",") + 
                     delimitedList(leByte | Literal("IG"))("reslist") +
                     SupLit(")")
                    )
# roomOps has sub-opcodes with non-standard encoding (mostly because of red, green, blue values)
# These are auto-generated... (I have modified some to make the question marks optional)
# A few sub-ops (RoomIntensity, SetScren) output an extra closing parenthesis
grFunc_roomOps_RoomScroll = Literal("RoomScroll")("function") + SupLit("(") + Group(leVarOrWord)("arg1") + SupLit(",") + Group(leVarOrWord)("arg2") + SupLit(")")
grFunc_roomOps_SetScreen = Literal("SetScreen")("function") + SupLit("(") + Group(leVarOrWord)("arg1") + SupLit(",") + Group(leVarOrWord)("arg2") + SupLit(")")
grFunc_roomOps_ShakeOn = Literal("ShakeOn")("function") + SupLit("(") + SupLit(")")
grFunc_roomOps_ShakeOff = Literal("ShakeOff")("function") + SupLit("(") + SupLit(")")
grFunc_roomOps_RoomIntensity = Literal("RoomIntensity")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(",") + Group(leVarOrByte)("arg2") + SupLit(",") + Group(leVarOrByte)("arg3") + SupLit(")")
grFunc_roomOps_saveLoad = (Literal("saveLoad?") ^ Literal("saveLoad"))("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(",") + Group(leVarOrByte)("arg2") + SupLit(")")
# screenEffect has a question mark and extra closing parenthesis
grFunc_roomOps_screenEffect = (Literal("screenEffect?") ^ Literal("screenEffect"))("function") + SupLit("(") + Group(leVarOrWord)("arg1") + SupLit(")")
grFunc_roomOps_colorCycleDelay = Literal("colorCycleDelay")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(",") + Group(leVarOrByte)("arg2") + SupLit(")")
# Uses aux opcode, also extra closing parenthesis
grFunc_roomOps_SetPalColor = (Literal("SetPalColor")("function") + SupLit("(") + 
                              Group(leVarOrWord)("red") + SupLit(",") + 
                              Group(leVarOrWord)("green") + SupLit(",") + 
                              Group(leVarOrWord)("blue") + SupLit(",") + 
                              Group(leVarOrByte)("index") +
                              SupLit(")")
                              )
# Lots of parameters, descumm calls it "Unused" which (while possibly true) isn't very descriptive
grFunc_roomOps_SetRoomScale = ((Literal("SetRoomScale") | Literal("Unused"))("function") + SupLit("(") + 
                         Group(leVarOrByte)("scale1") + SupLit(",") + 
                         Group(leVarOrByte)("y1") + SupLit(",") + 
                         Group(leVarOrByte)("scale2") + SupLit(",") + 
                         Group(leVarOrByte)("y2") + SupLit(",") + 
                         Group(leVarOrByte)("slot") + SupLit(",") + 
                         SupLit(")")
                         )
# This has params and an aux opcode, but I'm also cheating and using this for setRoomShadow.
grFunc_roomOps_setRGBRoomIntensity = ((Literal("setRGBRoomIntensity") | Literal("setRoomShadow"))("function") + SupLit("(") + 
                                      Group(leVarOrWord)("redscale") + SupLit(",") + 
                                      Group(leVarOrWord)("greenscale") + SupLit(",") + 
                                      Group(leVarOrWord)("bluescale") + SupLit(",") + 
                                      Group(leVarOrByte)("start") + SupLit(",") +
                                      Group(leVarOrByte)("end") +
                                      SupLit(")")
                                      )
# Share grammar & parse action for saveString and loadString
grFunc_roomOps_saveLoadString = ((Literal("saveString") | Literal("loadString"))("function") + SupLit("(") + 
                                 Group(leVarOrByte)("slot") + SupLit(",") + 
                                 leString("stringy") +
                                 SupLit(")")
                                 )
# Lots of params & aux opcodes
grFunc_roomOps_palManipulate = (Literal("palManipulate")("function") + SupLit("(") + 
                                Group(leVarOrByte)("slot") + SupLit(",") + 
                                Group(leVarOrByte)("start") + SupLit(",") + 
                                Group(leVarOrByte)("end") + SupLit(",") + 
                                Group(leVarOrByte)("time") +
                                SupLit(")"))
# This big thing marks the end of roomOps, phew! Also, looks like a lot of roomOps add
#  an extra closing parenthesis, so account for that here instead of in each sub-op.
grFunc_roomOps_SO = (grFunc_roomOps_RoomScroll | grFunc_roomOps_SetScreen | grFunc_roomOps_ShakeOn | grFunc_roomOps_ShakeOff | 
                     grFunc_roomOps_RoomIntensity | grFunc_roomOps_saveLoad | grFunc_roomOps_screenEffect | 
                     grFunc_roomOps_colorCycleDelay | grFunc_roomOps_SetPalColor | grFunc_roomOps_SetRoomScale |
                     grFunc_roomOps_setRGBRoomIntensity | grFunc_roomOps_saveLoadString | grFunc_roomOps_palManipulate) + Optional(SupLit(")"))
# setVarRange takes a starting variable, the number of constants that will be next in the instruction (I'm making this optional),
#  and a delmitedList of either bytes or words (but not mixed). We determine which in the parse action.
grFunc_setVarRange = (Literal("setVarRange")("function") + SupLit("(") +
                      Group(leVariable)("startvar") + SupLit(",") +
                      Optional(leByte("numitems") + SupLit(",")) +
                      SupLit("[") + delimitedList(leByte | leWord)("listvals") + SupLit("]") +
                      SupLit(")"))
# saveRestoreVerbs - I originally misunderstood this instruction; I thought it had an aux opcode, three params, then a sub-opcode,
#  when it actually just has a sub-opcode and three params. Still, this is more concise than the auto-generated instruction would be.
grFunc_saveRestoreVerbs_saveVerbs = Literal("saveVerbs")
grFunc_saveRestoreVerbs_restoreVerbs = Literal("restoreVerbs")
grFunc_saveRestoreVerbs_deleteVerbs = Literal("deleteVerbs")
grFunc_saveRestoreVerbs = ((grFunc_saveRestoreVerbs_saveVerbs | grFunc_saveRestoreVerbs_restoreVerbs | grFunc_saveRestoreVerbs_deleteVerbs)("function") + SupLit("(") +
                            Group(leVarOrByte)("start") + SupLit(",") +
                            Group(leVarOrByte)("end") + SupLit(",") +
                            Group(leVarOrByte)("mode") +
                            SupLit(")")
                            )
# startScript modifies the opcode, but descumm doesn't 
grFunc_startScript = (Literal("startScript")("function") + SupLit("(") +
                      Group(leVarOrByte)("arg1") + SupLit(",") +
                      leArgList("arg2") +
                      Optional(SupLit(",") + Literal("F"))("freezeres") + 
                      Optional(SupLit(",") + Literal("R"))("recursive") + 
                      SupLit(")"))
# stopObjectCode must be the last item in a script
grFunc_stopObjectCode = Literal("stopObjectCode")("function") + SupLit("(") + SupLit(")")
# systemOps takes sub-opcodes, but they're only represented as the numbers 1, 2, or 3. I also allow meaningful words.
grFunc_systemOps = (Literal("systemOps")("function") + SupLit("(") +
                    ((Literal("1") | Literal("RESTART")) | (Literal("2") | Literal("PAUSE")) | (Literal("3") | Literal("QUIT")))("subop") + # this is a bit crap
                    SupLit(")"))
# ... here endeth the quirky functions

# Pragramas (compiler directives)
grPragmaScriptNum = Literal("Script#") + leUnsignedInt("scriptnum")
grPragmaDefine = SupLit("#define") + leDefinedValueDef + Optional(SupLit("=")) + Group(leVarOrWord)
grPragma = grPragmaScriptNum | grPragmaDefine

# All quirky functions
grQuirkyFunctions = (grFunc_printEgo | grFunc_delay | grFunc_doSentence | grFunc_drawBox | grFunc_drawObject | grFunc_loadRoomWithEgo | grFunc_pickupObject |
                     grFunc_PseudoRoom | grFunc_oldRoomEffect | grFunc_override | grFunc_setVarRange | grFunc_saveRestoreVerbs | grFunc_systemOps | 
                     grFunc_roomOps_SO)

# Expression mode stuff
grSubExpression = Forward()
grExpressionMode = SupLit("Exprmode") + Group(leVariable)("target") + SupLit("=") + Group(grSubExpression)("subexp")
grExpressionOpcode = SupLit("<") + (grQuirkyFunctions | scummbler_autogram.grFunctions)("expop") + SupLit(">") # TODO: makes this only funcs that return a result
grSubExpression << (SupLit("(") + Group(leVarOrWord | grSubExpression | grExpressionOpcode)("arg1") + oneOf("+ - / * =")("operation") + Group(leVarOrWord | grSubExpression | grExpressionOpcode)("arg2") + SupLit(")"))

# Blocks and larger constructs
grConstruct = grIncDec | grInlineOperation | grExpressionMode | grJumpGoto | grJumpUnless | grQuirkyFunctions | scummbler_autogram.grFunctions # single-line construct
grLinePrefix = Optional(grJumpLabel) + Suppress(Optional(grDescummOpcode))
grLineSuffix = Suppress(Optional(leComment))

# grStartScriptLine is a hack to get around descumm not outputting the extra opcode flags
grStartScriptLine = Optional(grJumpLabel) + Optional(grDescummOpcode)("descummop") + grFunc_startScript + grLineSuffix
# grStopObjectCodeLine marks the end of a script
grStopObjectCodeLine = grLinePrefix + grFunc_stopObjectCode + grLineSuffix
grNormalLine = grLinePrefix + (grConstruct + Optional(SupLit(";")) | Suppress(leComment)) + grLineSuffix
grIfLineStart = grLinePrefix + grIfStart + grLineSuffix
grIfLineElse = grLinePrefix + grIfElse + (Optional(grIfStart)("elseif") ^ SupLit("{")) + grLineSuffix
grIfLineEnd = grLinePrefix + grIfEnd + grLineSuffix

grWhileLineStart = grLinePrefix + grWhileStart + grLineSuffix
grWhileLineEnd = grLinePrefix + grWhileEnd + grLineSuffix
grDoWhileLineStart = grLinePrefix + grDoWhileStart + grLineSuffix
grDoWhileLineEnd = grLinePrefix + grDoWhileEnd + grLineSuffix
grForLoopLineStart = grLinePrefix + grForLoopStart + grLineSuffix
grForLoopLineEnd = grLinePrefix + grForLoopEnd + grLineSuffix

grBlock = Forward()
# grBlock is optional, because descumm might try to be clever and create "if/else" statements when it's just doing
#  conditional jumps.
grIfBlock = grIfLineStart + Optional(OneOrMore(grBlock)) + Optional(OneOrMore(grIfLineElse + ZeroOrMore(grBlock))) + grIfLineEnd
grWhileBlock = grWhileLineStart + OneOrMore(grBlock) + grWhileLineEnd
grDoWhileBlock = grDoWhileLineStart + OneOrMore(grBlock) + grDoWhileLineEnd
grForLoopBlock = grForLoopLineStart + OneOrMore(grBlock) + grForLoopLineEnd
grBlock << (grNormalLine ^ grStartScriptLine ^ grIfBlock ^ grDoWhileBlock ^ grWhileBlock ^ grForLoopBlock)
grMainBlock = StringStart() + ZeroOrMore(grPragma) + ZeroOrMore(grBlock) + grStopObjectCodeLine + Optional(SupLit("END")) + StringEnd()
# This is used for testing, so we don't have to put "stopObjectCode()" at the end of every test
grMainBlockTesting = StringStart() + ZeroOrMore(grPragma) + ZeroOrMore(grBlock) + Optional(grStopObjectCodeLine) + Optional(SupLit("END")) + StringEnd()

# } End Grammar


#__________________
# Start Parse Actions
    
def do_grInlineOperation(s, loc, toks):
    op = opInlineOperation[toks.operation]
    val, op = resolve_parameter(toks.value, op, 0x80, RP_WORD)
    targ = resolve_var(toks.target)
    
    return to_byte(op) + targ + val

def do_grIncDec(s, loc, toks):
    op = opInlineOperation[toks.operation]
    targ = resolve_var(toks.target)
    
    return to_byte(op) + targ

def do_grExpressionMode(s, loc, toks):
    op = opExpressionOpcode
    targ = resolve_var(toks.target)
    subexp = toks.subexp

    return to_byte(op) + targ + subexp[0] + to_byte(opExpressionEnd)

def do_grSubExpression(s, loc, toks):
    argop1 = opExpressionValue # these come from scummbler_opcodes
    argop2 = opExpressionValue
    stackop = opExpressionOperation[toks.operation]

    omitfirst = False
    omitsecond = False
    
    # These are some really crappy workarounds here, I probably need to revise my grammar
    if 'expop' in toks.arg1: # sub-instruction
        val1 = toks.arg1[0]
        argop1 = opExpressionSubOpcode
    elif not 'vartype' in toks.arg1 and not 'knownvar' in toks.arg1: # assume it's a constant or subexpression
        if not 'value' in toks.arg1: # subexpression
            val1 = toks.arg1[0]
            # Need to omit first argop as subexpr already has it
            omitfirst = True
        else: # constant
            val1 = to_word_LE(toks.arg1.value)
    else:
        val1 = resolve_var(toks.arg1)
        argop1 = argop1 | 0x80 # inline operations only take one parameter
    
    if 'expop' in toks.arg2: # sub-instruction
        val2 = toks.arg2[0]
        argop2 = opExpressionSubOpcode
    elif not 'vartype' in toks.arg2 and not 'knownvar' in toks.arg2: # assume it's a constant or subexpression
        if not 'value' in toks.arg2: # subexpression
            val2 = toks.arg2[0]
            # Need to omit first argop as subexpr already has it
            omitsecond = True
        else: # constant
            val2 = to_word_LE(toks.arg2.value)
    else:
        val2 = resolve_var(toks.arg2)
        argop2 = argop2 | 0x80 # inline operations only take one parameter
        
    return ("" if omitfirst else to_byte(argop1)) + val1 + \
            ("" if omitsecond else to_byte(argop2)) + val2 + \
            to_byte(stackop)

def do_grConstruct(s, loc, toks):
    # This stores the line numbers for meaningful constructs. Avoids problems
    #  with blank lines, comments, etc. Does not affect tokens at all.
    global lineNum
    lineNum += 1

def do_grJumpGoto(s, loc, toks):
    global jumpTable
    global lineNum
    if not jumpTable.has_key(toks.target):
        jumpTable[toks.target] = []
    jumpTable[toks.target].append(lineNum)
    return "\x18\x00\x00" # placeholder, will modify this in Phase 2

def do_grJumpLabel(s, loc, toks):
    # Labels are only used by the compiler; returns nothing.
    global lineNum
    global labelTable
    if labelTable.has_key(toks.label):
        if labelTable[toks.label] == lineNum:
            # If we're parsing a label we've already tried parsing (could occur with back-tracking),
            #  just ignore the label
            return []
        raise ParseFatalException("Duplicate label declaration on line " + str(lineno(loc, s)) + ": " + toks.label)
    labelTable[toks.label] = lineNum
    return []

def do_grIfZeroEquality(s, loc, toks):
    if not '!' in toks:
        op = opComparisonsNoParameters["notEqualZero"]
    else:
        op = opComparisonsNoParameters["equalZero"]

    if not 'vartype' in toks and not 'knownvar' in toks: # should never happen
        raise ParseFatalException("The parameter in a \"equalZero\" or \"notEqualZero\" test must be a variable: " + line(loc, s))
    else:
        val1 = resolve_var(toks)
    
    return to_byte(op) + val1

def do_grIfComparison(s, loc, toks):
    op = opComparisons[toks.comparator]
    
    val1 = resolve_var(toks.arg1)
    
    # Because arg1 is always a variable, parameter bits tell what arg2 is
    val2, op = resolve_parameter(toks.arg2, op, 0x80, RP_WORD)
    
    return to_byte(op) + val1 + val2

def do_grIfClassOfIs(s, loc, toks):
    op = opFunctionTable[toks.function]
    testobject, op = resolve_parameter(toks.testobject, op, 0x80, RP_WORD)
    classes = ""
    for c in toks.classes:
        aux = 0x01
        c, aux = resolve_parameter(c, aux, 0x80, RP_WORD)
        classes += to_byte(aux) + c
    classes += "\xFF"
    return to_byte(op) + testobject + classes

def do_grIfActorInBox(s, loc, toks):
    op = opFunctionTable[toks.function]
    actor, op = resolve_parameter(toks.actor, op, 0x80, RP_BYTE)
    box, op = resolve_parameter(toks.box, op, 0x40, RP_BYTE)
    return to_byte + actor + box

def do_grIfState(s, loc, toks):
    if toks.comparator == "==":
        op = opFunctionTable["ifState"]
    else: # toks.comparator == "!="
        op = opFunctionTable["ifNotState"]
    testobject, op = resolve_parameter(toks.testobject, op, 0x80, RP_WORD)
    state, op = resolve_parameter(toks.state, op, 0x40, RP_BYTE)
    return to_byte(op) + testobject + state

def do_grJumpUnless(s, loc, toks):
    """ Similar to do_grIfLineStart, but without keeping track of "if" blocks etc.
    Also similar to do_grJumpGoto, but we use the already parsed operator etc."""
    global jumpTable
    global lineNum
    if not jumpTable.has_key(toks.target):
        jumpTable[toks.target] = []
    jumpTable[toks.target].append(lineNum)
    return toks[0] + "\x00\x00"
    
def do_grIfLineStart(s, loc, toks):
    # Seperate parsing of "if" statements & lines; "if" could be either a new "if" block
    #  or part of an "else if" statement. This method allows us to create a stack for
    #  nested if blocks.
    global lineNum
    global ifStack
    global currIfBlock

    # Push current "if" block data onto the stack so we can come back to it if
    #  necessary. This means the first entry will always be a dummy value of our
    #  initial settings.
    ifStack.append( currIfBlock )
    currIfBlock = IfBlockInfo(lineNum)
    
    lineNum += 1
    return toks[0] + "\x00\x00" # most parsing was already done in grIfStart, we just add the jump
    
def do_grIfElse(s, loc, toks):
    global lineNum
    global currIfBlock
    
    # Push this onto the branch stack to be resolved when we find the "if" end.
    # This is a goto at the end of the block we just finished parsing.
    currIfBlock.pushBranch(lineNum)
    
    lineNum += 1
    return "\x18\x00\x00"
    
def do_grIfLineElse(s, loc, toks):
    # Similar to do_grIfLineStart, but without manipulating the ifStack.
    global lineNum
    global currIfBlock
    global jumpTable
    global labelTable # ^ these two should be merged
    
    # lineNum has already been adjusted to be after the goto inserted by the "else".
    # (pop -2 because we just added a goto and pushed its details)
    lastLine, lastLabel = currIfBlock.popElseBranch()

    jumpTable[lastLabel] = [lastLine]
    labelTable[lastLabel] = lineNum

    if toks.elseif != "":
        currIfBlock.pushBranch(lineNum)
        lineNum += 1
        toks[1] = toks[1] + "\x00\x00" # add placeholder for jump
    
    return toks
    
def do_grIfEnd(s, loc, toks):
    global lineNum
    global ifStack
    global currIfBlock
    global jumpTable
    global labelTable
    
    # Map all remaining branch jumps to this end position.
    for lastLine, lastLabel in currIfBlock:
        jumpTable[lastLabel] = [lastLine]
        labelTable[lastLabel] = lineNum
    
    currIfBlock = ifStack.pop()
    
    # Note that the original SCUMM compiler emits dummy jumps at the end
    #  of every "else if" block. Scummbler does not, as it works on a
    #  lower level, handles "if" blocks differently, and relies on
    #  output from descumm, which outputs dummy jumps as comments.
    
    return []

def do_grWhileStart(s, loc, toks):
    global loopStack
    global lineNum
    
    loopStack.append(LoopBlockInfo(lineNum))
    lineNum += 1
    return toks[0] + "\x00\x00"

def do_grWhileEnd(s, loc, toks):
    global loopStack
    global lineNum
    
    loopinfo = loopStack.pop()
    loopStart = loopinfo.startLine
    
    # Generate loop back to start
    label = gen_label('w', loopStart, 0) # 0 marks the beginning of loops
    labelTable[label] = loopStart
    if not jumpTable.has_key(label):
        jumpTable[label] = []
    jumpTable[label].append(lineNum)
    
    # Generate jump from start to end of block
    label = gen_label('w', loopStart, 1) # 1 marks the end of loops
    labelTable[label] = lineNum
    if not jumpTable.has_key(label):
        jumpTable[label] = []
    jumpTable[label].append(loopStart)
    
    lineNum += 1
    
    return "\x18\x00\x00"    

def do_grDoWhileStart(s, loc, toks):
    global lineNum
    global loopStack
    
    # Just mark the beginning of the loop
    loopStack.append(LoopBlockInfo(lineNum))
    
    return []

def do_grDoWhileEnd(s, loc, toks):
    global loopStack
    global lineNum
    
    # Generate loop back to start
    loopStart = loopStack.pop().startLine
    label = gen_label('d', loopStart, 0) # 0 marks the beginning of loops
    labelTable[label] = loopStart
    if not jumpTable.has_key(label):
        jumpTable[label] = []
    jumpTable[label].append(lineNum)
    
    lineNum += 1
    return toks[0] + "\x00\x00"

def do_grForLoopStart(s, loc, toks):
    global loopStack
    global lineNum
    
    output = []
    
    if 'init' in toks:
        output.append(toks.init[0])    
        lineNum += 1
    
    output.append(toks.test[0] + '\x00\x00')

    increm = None
    if 'increm' in toks:
        increm = toks.increm
        
    loopStack.append(LoopBlockInfo(lineNum, increm))
        
    lineNum += 1
    
    return output

def do_grForLoopEnd(s, loc, toks):
    # Similar to do_grWhileEnd, but adds the increment instruction
    global loopStack
    global lineNum

    output = []
    
    loopinfo = loopStack.pop()
    loopStart = loopinfo.startLine
    
    # Add "increment" instruction if present
    if loopinfo.incremOp != None:
        output.append(loopinfo.incremOp)
        lineNum += 1
    
    # Generate loop back to start
    label = gen_label('w', loopStart, 0) # 0 marks the beginning of loops
    labelTable[label] = loopStart
    if not jumpTable.has_key(label):
        jumpTable[label] = []
    jumpTable[label].append(lineNum)
    
    lineNum += 1
    
    # Generate jump from start to end of block
    label = gen_label('w', loopStart, 1) # 1 marks the end of loops
    labelTable[label] = lineNum
    if not jumpTable.has_key(label):
        jumpTable[label] = []
    jumpTable[label].append(loopStart)

    output.append("\x18\x00\x00")
    
    return output
    

# --- Function parse actions ---
# NOTE: these have hardcoded opcodes, which is very naughty!
def do_grFunc_delay(s, loc, toks):
    op = opFunctionTable[toks.function]
    val = int(toks.arg1[0])
    return to_byte(op) + chr(val & 0xFF) + chr((val & 0xFF00) >> 8) + chr((val & 0xFF0000) >> 16) 

def do_grFunc_doSentence(s, loc, toks):
    target = arg1 = arg2 = arg3 = ""
    op = opFunctionTable[toks.function]
    # No target
    if (toks.arg1 == "STOP"):
        return to_byte(op) + to_byte(0xFE)
    arg1, op = resolve_parameter(toks.arg1, op, 0x80, False)
    arg2, op = resolve_parameter(toks.arg2, op, 0x40, True)
    arg3, op = resolve_parameter(toks.arg3, op, 0x20, True)
    return to_byte(op) + target + arg1 + arg2 + arg3

def do_grFunc_drawBox(s, loc, toks):
    op = opFunctionTable[toks.function]
    
    left, op = resolve_parameter(toks.left, op, 0x80, RP_WORD)
    top, op = resolve_parameter(toks.top, op, 0x40, RP_WORD)
    
    auxop = 0x01 # not sure if this auxop is actually 0x01
    
    right, op = resolve_parameter(toks.right, auxop, 0x80, RP_WORD)
    bottom, op = resolve_parameter(toks.bottom, auxop, 0x40, RP_WORD)
    colour, op = resolve_parameter(toks.colour, auxop, 0x20, RP_BYTE)
    
    return to_byte(op) + left + top + to_byte(auxop) + right + bottom + colour

def do_grFunc_drawObject_setXY(s, loc, toks):
    target = arg1 = arg2 = arg3 = ""
    op = opFunctionTable[toks.function]
    # No target
    arg1, op = resolve_parameter(toks.arg1, op, 0x80, RP_WORD)
    arg2, op = resolve_parameter(toks.arg2, op, 0x40, RP_WORD)
    # No arg3
    return to_byte(op) + target + arg1 + arg2 + arg3

def do_grFunc_drawObject_setImage(s, loc, toks):
    target = arg1 = arg2 = arg3 = ""
    op = opFunctionTable[toks.function]
    # No target
    arg1, op = resolve_parameter(toks.arg1, op, 0x80, RP_WORD)
    # No arg2
    # No arg3
    return to_byte(op) + target + arg1 + arg2 + arg3

def do_grFunc_drawObject(s, loc, toks): # quirky function
    target = arg1 = arg2 = arg3 = ""
    op = opFunctionTable[toks.function]
    # No target
    arg1, op = resolve_parameter(toks.arg1, op, 0x80, RP_WORD)
    if not 'arg2' in toks:
        arg2 = to_byte(31) # compensate for "empty" sub-opcode
    else:
        arg2 = ''.join(toks.arg2) # I can't remember why we do this
    # No arg3
    return to_byte(op) + target + arg1 + arg2 + arg3

def do_grFunc_loadRoomWithEgo(s, loc, toks):
    target = arg1 = arg2 = arg3 = ""
    op = opFunctionTable[toks.function]
    # No target
    objecttok, op = resolve_parameter(toks.object, op, 0x80, RP_WORD)
    room, op = resolve_parameter(toks.room, op, 0x40, RP_BYTE)
    x = to_word_LE(toks.x)
    y = to_word_LE(toks.y)
    return to_byte(op) + objecttok + room + x + y

def do_grFunc_oldRoomEffect(s, loc, toks):
    op = opFunctionTable[toks.function]
    # I'm not sure if these values are okay...
    if toks.function == "oldRoomEffect-set":
        subop = 0x03
    else: # oldRoomEffect-fadein
        subop = 0x01
    effect, op = resolve_parameter(toks.effect, op, 0x80, RP_WORD)
    return to_byte(op) + to_byte(subop) + effect

def do_grFunc_override(s, loc, toks):
    return to_byte(88) + to_byte(0x01 if toks.function == "beginOverride" else 0x00)

def do_grFunc_pickupObject(s, loc, toks):
    target = arg1 = arg2 = arg3 = ""
    if "arg2" in toks:
        op = opFunctionTable[toks.function]
    else:
        op = opFunctionTable["pickupObjectOld"]
    # No target
    arg1, op = resolve_parameter(toks.arg1, op, 0x80, RP_WORD)
    if "arg2" in toks:
        arg2, op = resolve_parameter(toks.arg2, op, 0x40, RP_BYTE)
    # No arg3
    return to_byte(op) + target + arg1 + arg2 + arg3

def do_grFunc_print(s, loc, toks):
    arg1 = arg2 = arg3 = ""
    op = opFunctionTable[toks.function]
    # No target
    arg1, op = resolve_parameter(toks.arg1, op, 0x80, False)
    if not 'arg2' in toks: # workaround for no sub-opcodes
        arg2 = ''
        arg3 = '\xFF'
    else:
        arg2 = ''.join(toks.arg2)
        # print will end with 0x00 if it displays text (sub-op 0x0F),
        # otherwise it terminates with 0xFF.
        if toks.arg2[-1][0] != "\x0F":
            arg3 = "\xFF"
    return to_byte(op) + arg1 + arg2 + arg3

def do_grFunc_printEgo(s, loc, toks):
    arg1 = arg2 = ""
    op = opFunctionTable[toks.function]
    if not 'arg1' in toks: # workaround for no sub-opcodes
        arg1 = ''
        arg2 = '\xFF'
    else:
        arg1 = ''.join(toks.arg1)
        # print will end with 0x00 if it displays text (sub-op 0x0F),
        # otherwise it terminates with 0xFF.
        if toks.arg1[-1][0] != "\x0F":
            arg2 = "\xFF"
    return to_byte(op) + arg1 + arg2

def do_grFunc_PseudoRoom(s, loc, toks):
    op = opFunctionTable[toks.function]
    val = to_byte(toks.val)
    resvals = ""
    for r in toks.reslist:
        if r == "IG":
            r = "0"
        resvals += to_byte(r, 0x80)
    resvals += "\x00"
    return to_byte(op) + val + resvals

# Auto-generated roomOps sub-opcodes
def do_grFunc_roomOps_RoomScroll(s, loc, toks):
    target = arg1 = arg2 = arg3 = ""
    op = opFunctionTable[toks.function]
    # No target
    arg1, op = resolve_parameter(toks.arg1, op, 0x80, RP_WORD)
    arg2, op = resolve_parameter(toks.arg2, op, 0x40, RP_WORD)
    # No arg3
    return to_byte(op) + target + arg1 + arg2 + arg3

def do_grFunc_roomOps_SetScreen(s, loc, toks):
    target = arg1 = arg2 = arg3 = ""
    op = opFunctionTable[toks.function]
    # No target
    arg1, op = resolve_parameter(toks.arg1, op, 0x80, RP_WORD)
    arg2, op = resolve_parameter(toks.arg2, op, 0x40, RP_WORD)
    # No arg3
    return to_byte(op) + target + arg1 + arg2 + arg3

def do_grFunc_roomOps_ShakeOn(s, loc, toks):
    target = arg1 = arg2 = arg3 = ""
    op = opFunctionTable[toks.function]
    # No target
    # No arg1
    # No arg2
    # No arg3
    return to_byte(op) + target + arg1 + arg2 + arg3

def do_grFunc_roomOps_ShakeOff(s, loc, toks):
    target = arg1 = arg2 = arg3 = ""
    op = opFunctionTable[toks.function]
    # No target
    # No arg1
    # No arg2
    # No arg3
    return to_byte(op) + target + arg1 + arg2 + arg3

def do_grFunc_roomOps_RoomIntensity(s, loc, toks):
    target = arg1 = arg2 = arg3 = ""
    op = opFunctionTable[toks.function]
    # No target
    arg1, op = resolve_parameter(toks.arg1, op, 0x80, RP_BYTE)
    arg2, op = resolve_parameter(toks.arg2, op, 0x40, RP_BYTE)
    arg3, op = resolve_parameter(toks.arg3, op, 0x20, RP_BYTE)
    return to_byte(op) + target + arg1 + arg2 + arg3

def do_grFunc_roomOps_saveLoad(s, loc, toks):
    target = arg1 = arg2 = arg3 = ""
    op = opFunctionTable[toks.function]
    # No target
    arg1, op = resolve_parameter(toks.arg1, op, 0x80, RP_BYTE)
    arg2, op = resolve_parameter(toks.arg2, op, 0x40, RP_BYTE)
    # No arg3
    return to_byte(op) + target + arg1 + arg2 + arg3

def do_grFunc_roomOps_screenEffect(s, loc, toks):
    target = arg1 = arg2 = arg3 = ""
    op = opFunctionTable[toks.function]
    # No target
    arg1, op = resolve_parameter(toks.arg1, op, 0x80, RP_WORD)
    # No arg2
    # No arg3
    return to_byte(op) + target + arg1 + arg2 + arg3

def do_grFunc_roomOps_colorCycleDelay(s, loc, toks):
    target = arg1 = arg2 = arg3 = ""
    op = opFunctionTable[toks.function]
    # No target
    arg1, op = resolve_parameter(toks.arg1, op, 0x80, RP_BYTE)
    arg2, op = resolve_parameter(toks.arg2, op, 0x40, RP_BYTE)
    # No arg3
    return to_byte(op) + target + arg1 + arg2 + arg3

# Manual roomOps sub-opcodes
def do_grFunc_roomOps_SetPalColor(s, loc, toks):
    op = opFunctionTable[toks.function]
    red, op = resolve_parameter(toks.red, op, 0x80, RP_WORD)
    green, op = resolve_parameter(toks.green, op, 0x40, RP_WORD)
    blue, op = resolve_parameter(toks.blue, op, 0x20, RP_WORD)
    auxop = 0x01 # need to verify this
    index, auxop = resolve_parameter(toks.index, auxop, 0x80, RP_WORD)
    return to_byte(op) + red + green + blue + to_byte(auxop) + index

def do_grFunc_roomOps_SetRoomScale(s, loc, toks):
    op = opFunctionTable[toks.function]
    scale1, op = resolve_parameter(toks.scale1, op, 0x80, RP_BYTE)
    y1, op = resolve_parameter(toks.y1, op, 0x40, RP_BYTE)
    auxop1 = 0x01
    scale2, auxop1 = resolve_parameter(toks.scale2, auxop1, 0x80, RP_BYTE)
    y2, auxop1 = resolve_parameter(toks.y2, auxop1, 0x40, RP_BYTE)
    auxop2 = 0x01
    slot, auxop2 = resolve_parameter(toks.slot, auxop2, 0x80, RP_BYTE)
    return to_byte(op) + scale1 + y1 + to_byte(auxop1) + scale2 + y2 + to_byte(auxop2) + slot

def do_grFunc_roomOps_setRGBRoomIntensity(s, loc, toks):
    opFunctionTable[toks.function]
    redscale, op = resolve_parameter(toks.redscale, op, 0x80, RP_WORD)
    greenscale, op = resolve_parameter(toks.greenscale, op, 0x40, RP_WORD)
    bluescale, op = resolve_parameter(toks.bluescale, op, 0x20, RP_WORD)
    auxop = 0x01 # TODO: verify this
    start, auxop = resolve_parameter(toks.start, auxop, 0x80, RP_BYTE)
    end, auxop = resolve_parameter(toks.end, auxop, 0x40, RP_BYTE)
    return to_byte(op) + redscale + greenscale + bluescale + to_byte(auxop) + start + end

def do_grFunc_roomOps_saveLoadString(s, loc, toks):
    opFunctionTable[toks.function]
    slot, op = resolve_parameter(toks.slot, op, 0x80, RP_BYTE)
    stringy = toks.stringy + "\x00"
    return to_byte(op) + slot + stringy

def do_grFunc_roomOps_palManipulate(s, loc, toks):
    opFunctionTable[toks.function]
    slot, op = resolve_parameter(toks.slot, op, 0x80, RP_BYTE)
    auxop1 = 0x01
    start, auxop1 = resolve_parameter(toks.start, auxop1, 0x80, RP_BYTE)
    end, auxop1 = resolve_parameter(toks.end, auxop1, 0x40, RP_BYTE)
    auxop2 = 0x01
    time, auxop2 = resolve_parameter(toks.time, auxop2, 0x80, RP_BYTE)
    return to_byte(op) + slot + to_byte(auxop1) + start + end + to_byte(auxop2) + time

def do_grFunc_roomOps_SO(s, loc, toks):
    return to_byte(opFunctionTable["roomOps"]) + ''.join(toks) # bit dodgy

def do_grFunc_saveRestoreVerbs(s, loc, toks):
    op = opFunctionTable[toks.function]
    if toks.function == "saveVerbs":
        subop = 0x01
    elif toks.function == "restoreVerbs":
        subop = 0x02
    elif toks.function == "deleteVerbs":
        subop = 0x03
    else: # will never happen
        raise ScummblerParseException("What the blooming heck have you done, my lad? " + str(toks.function) + " is not saveRestoreVerbs-ish!")
    
    start, subop = resolve_parameter(toks.start, subop, 0x80, RP_BYTE)
    end, subop = resolve_parameter(toks.end, subop, 0x40, RP_BYTE)
    mode, subop = resolve_parameter(toks.mode, subop, 0x20, RP_BYTE)
    
    return to_byte(op) + to_byte(subop) + start + end + mode

def do_grFunc_setVarRange(s, loc, toks):
    op = opFunctionTable[toks.function]
    startvar = resolve_var(toks.startvar)
    if "numitems" in toks:
        numitems = int(toks.numitems)
    else:
        numitems = None
    vals = []
    use_words = False
    # If highest (or lowest negative) value can't fit in a byte,
    #  we'll use words!
    # We could potentially check for negatives, set a boolean if found,
    #  and then check between -128 and 127.
    for v in toks.listvals:
        v = int(v)
        if v < -128 or v > 255:
            use_words = True
            op = op | 0x80
        vals.append(v)
        
    if numitems is None:
        numitems = len(vals)
    elif numitems != len(vals):
        raise ScummblerParseException("setVarRange: the given number of values does not match the actual number of values.")
        
    if use_words:
        vals = ''.join(to_word_LE(v) for v in vals)
    else:
        vals = ''.join(to_byte(v) for v in vals)
    
    return to_byte(op) + startvar + to_byte(numitems) + vals

def do_grFunc_startScript(s, loc, toks):
    global lineNum
    arg1 = arg2 = arg3 = ""
    op = opFunctionTable[toks.function]
    # No target
    arg1, op = resolve_parameter(toks.arg1, op, 0x80, False)
    for a in toks.arg2:
        aux = 0x01
        a, aux = resolve_parameter(a, aux, 0x80, True)
        arg2 += to_byte(aux) + a
    arg2 += "\xFF"
    
    # Explicit flag decleration (can be overriden by existing input opcde)
    if "recursive" in toks:
        op |= 0x40
    if "freezeres" in toks:
        op |= 0x20
    
    lineNum += 1
    return to_byte(op) + arg1 + arg2 + arg3

def do_grStartScriptLine(s, loc, toks):
    # NOTE: this could screw up explicit flag declaration
    if "descummop" in toks:
        flags = int("0x" + toks[0], 16) & 0x60
        #toks[1][0] = to_byte(ord(toks[1][0]) | flags)
        return to_byte(ord(toks[1][0]) | flags) + toks[1][1:]
    return toks

def do_grFunc_stopObjectCode(s, loc, toks):
    op = opFunctionTable[toks.function]
    return to_byte(op)

def do_grFunc_systemOps(s, loc, toks):
    op = opFunctionTable[toks.function]
    subop = toks.subop
    if toks.subop == "1" or toks.subop == "RESTART":
        subop = 0x01
    elif toks.subop == "2" or toks.subop == "PAUSE":
        subop = 0x02
    elif toks.subop == "3" or toks.subop == "QUIT":
        subop = 0x03
    else: # will never happen
        raise ScummblerParseException("What the blooming heck have you done, my lad? " + str(toks.function) + " is not systemOps-ish!")
    
    return to_byte(op) + to_byte(subop)

def do_grPragmaScriptNum(s, loc, toks):
    # Because this returns some code, we need to increment the line counter
    #  so jumps resolve correctly.
    global lineNum
    global scriptType
    scriptType = "LSCR" # for now we just assume if this script is numbered, it's local
    lineNum += 1 # because this outputs a value we need to increment lineNum
    return to_byte(toks.scriptnum)

def do_grPragmaDefine(s, loc, toks):
    global definedValues
    definedValues[toks[0]] = toks[1]
    return []

# My original intention for this was different grammars could share parse
#  actions, but that didn't end up being useful. This list could be automated.
actionMap = {
    grInlineOperation : do_grInlineOperation,
    grIncDec : do_grIncDec,
    grExpressionMode : do_grExpressionMode,
    grSubExpression : do_grSubExpression,
    grJumpLabel : do_grJumpLabel,
    grJumpGoto : do_grJumpGoto,
    grConstruct : do_grConstruct,
    grIfZeroEquality : do_grIfZeroEquality,
    grIfComparison : do_grIfComparison,
    grIfClassOfIs : do_grIfClassOfIs,
    grIfActorInBox : do_grIfActorInBox,
    grIfState : do_grIfState,
    grJumpUnless : do_grJumpUnless,
    grIfLineStart : do_grIfLineStart,
    grIfElse : do_grIfElse,
    grIfLineElse : do_grIfLineElse,
    grIfEnd : do_grIfEnd,
    grWhileStart : do_grWhileStart,
    grWhileEnd : do_grWhileEnd,
    grDoWhileStart : do_grDoWhileStart,
    grDoWhileEnd : do_grDoWhileEnd,
    grForLoopStart : do_grForLoopStart,
    grForLoopEnd : do_grForLoopEnd,
    grFunc_delay : do_grFunc_delay,
    grFunc_doSentence : do_grFunc_doSentence,
    grFunc_drawBox : do_grFunc_drawBox,
    grFunc_drawObject_setXY : do_grFunc_drawObject_setXY,
    grFunc_drawObject_setImage : do_grFunc_drawObject_setImage,
    grFunc_drawObject : do_grFunc_drawObject,
    grFunc_loadRoomWithEgo : do_grFunc_loadRoomWithEgo,
    grFunc_pickupObject : do_grFunc_pickupObject,
    scummbler_autogram.grFunc_print : do_grFunc_print,
    grFunc_printEgo : do_grFunc_printEgo,
    grFunc_PseudoRoom : do_grFunc_PseudoRoom,
    grFunc_roomOps_RoomScroll : do_grFunc_roomOps_RoomScroll,
    grFunc_roomOps_SetScreen : do_grFunc_roomOps_SetScreen,
    grFunc_roomOps_ShakeOn : do_grFunc_roomOps_ShakeOn,
    grFunc_roomOps_ShakeOff : do_grFunc_roomOps_ShakeOff,
    grFunc_roomOps_RoomIntensity : do_grFunc_roomOps_RoomIntensity,
    grFunc_roomOps_saveLoad : do_grFunc_roomOps_saveLoad,
    grFunc_roomOps_screenEffect : do_grFunc_roomOps_screenEffect,
    grFunc_roomOps_colorCycleDelay : do_grFunc_roomOps_colorCycleDelay,
    grFunc_roomOps_SetPalColor : do_grFunc_roomOps_SetPalColor,
    grFunc_roomOps_SetRoomScale : do_grFunc_roomOps_SetRoomScale,
    grFunc_roomOps_setRGBRoomIntensity : do_grFunc_roomOps_setRGBRoomIntensity,
    grFunc_roomOps_saveLoadString : do_grFunc_roomOps_saveLoadString,
    grFunc_roomOps_palManipulate : do_grFunc_roomOps_palManipulate,
    grFunc_roomOps_SO : do_grFunc_roomOps_SO,
    grFunc_setVarRange : do_grFunc_setVarRange,
    grFunc_saveRestoreVerbs : do_grFunc_saveRestoreVerbs,
    grFunc_startScript : do_grFunc_startScript,
    grStartScriptLine : do_grStartScriptLine,
    grFunc_stopObjectCode : do_grFunc_stopObjectCode,
    grFunc_systemOps : do_grFunc_systemOps,
    grFunc_oldRoomEffect : do_grFunc_oldRoomEffect,
    grFunc_override : do_grFunc_override,
    grFunc_stopObjectCode : do_grFunc_stopObjectCode,
    grPragmaScriptNum : do_grPragmaScriptNum,
    grPragmaDefine : do_grPragmaDefine
}

# End Parse Actions


def resolve_jumps(results):
    global jumpTable
    global labelTable

    # For each label being jumped to, resolve each jump to that label.
    # (labels without jumps to them are ignored)
    # l = the line number containing the label
    # j = the line number containing the jump
    for label, jlist in jumpTable.items():
        if not labelTable.has_key(label):
            raise ScummblerException("Label does not exist: " + str(label) + "  (affected meaningful lines: " + str(jlist) + ")")
        
        l = labelTable[label] # line number of the label
        
        for j in jlist:    
            if l <= j:
                offset = -sum( [ len(results[i]) for i in xrange(l, j + 1) ] )
            else:
                # Also handles jump to next line (offset of 0)
                # I'm not going to optimize 0-offset jumps because "descumm" will expect
                #  unoptimized code, and if the useless gotos are missing, it might
                #  just very well crawl into a ditch and die.
                offset = sum( [ len(results[i]) for i in xrange(j + 1, l) ] )
                
            if offset < minJump or offset > maxJump:
                raise ScummblerException("Jump is too large: " + str(offset) + " (meaningful line: " + str(j) + ")")
            
            results[j] = results[j][:-2] + to_word_LE(offset)



def do_test():
    global jumpTable
    global labelTable
    global actionMap
    global lineNum
    
    testlist = [
"Local[0] += 7", # Test 01: addition
"Var[166 + Var[165]] = Var[310 + Local[0]]", # Test 02: indirect
"Local[15] = Var[64 + 22]", # Test 03: indirect
"""Bit[22]--
Var[230]++;
Var[3 + Local[11]] &= Var[4 + Var[24]]""", # Test 04: inc, dec, indirect anding
"""[002B] (5A) Var[229] += 8;
[00A9] Var[229] += 8
(5A) Var[229] += Var[230];""", # Test 05: labels, descummed crud
"""Var[256] /= Bit[254] /* comment 1 */
/* This line is a comment */
[005D] /* Var[155]++
[006F] Local[2] -= Var[155]*/""", # Test 06: comments
"Local[2] = Gibberish", # Test 07: expected parse failure
"Exprmode Local[2] = (Var[238] + 6);", # Test 08: expression mode
"Exprmode Var[100] = ((120 + Var[165]) - 1);", # Test 09: expression mode (nested subexpression)
"""
goto 006F;
/* FLARGH */

[00A9] Var[229] += 8
[006F] Local[2] -= Var[155]

goto 00A9;
goto infloop;
[infloop] goto infloop

""", # Test 10: empty lines, jumps
"""[0013] (C4) if (Local[2] > Var[308]) {
Local[2] += 8
}
[afterif] /* hi there */
Local[0]++
goto afterif""", # Test 11: if statement
"""[0013] if (Local[2] > Var[308]) {
Local[2] += 8;
} else if (Local[2] > Var[308]) {
Local[3] += 4;
} else {
Local[7] += 5
}

[infloop] goto infloop""", # Test 12: if/else statement
"""
[0013] (C4) if (Local[2] > Var[308]) {
[001A] (AC)   Exprmode Var[238] = (Var[308] - 6);
[0025] (78)   if (Var[238] < 0) {
[002C] (1A)     Var[238] = 0;
[0031] (**)   }
[0031] (**) }
""", # Test 13: real if/else example. Nested if statements.
"""
[monkey] Var[238] = 0;
if (Local[7]) {
    unless (!Var[238]) goto monkey;
}
Var[238] = 1
""", # Test 14: equality to zero tests, "unless" statement.
"""
breakHere();
Var[240] = actorFromPos(-8,8)
walkActorTo(Var[1],179,44)
setObjectName(124,"bucket o' mud")
cutscene([2])
chainScript(3,[Local[0],Local[1]]);
setClass(346, [1,2,3,4,Var[5]])
startScript(11,[])
""", # Test 15: functions, auto-generated grammar, a negative number (wrong usage), varargs
"""
ActorOps(11,[TalkAnimNr(4,5)]);
ActorOps(8,[Init(),Costume(21),IgnoreBoxes(),SetZClip(1)]);
ActorOps(Var[1],[Init(),Costume(1),TalkColor(15),Name("Guybrush")]);
ActorOps(8,[Init(),Costume(28),IgnoreBoxes(),NeverZClip(),WalkAnimNr(7),StandAnimNr(7),InitAnimNr(7),WalkSpeed(15,15)]);
ActorOps(Var[1],[TalkColor(Var[485])]);
""", # Test 16: functions with sub-opcodes (multival, emit root)
"""
CursorShow()
SetCursorImg(24,25)
CursorCommand([6,7,8,9])
createBoxMatrix()
""", # Test 17: functions with sub-opcodes (non-multival, omit root), varargs
"""
printEgo([Text("What's that?")]);
print(252,[Color(12),Text("Are you sure you want to win? (Y/N)")]);
printEgo([Text("Nice ^255^6^1@.")]);
print(6,[Text("Well, you'd better have more tomorrow^")]);
""", # Test 18: functions with sub-opcodes (print), quirky function, escape codes & "..." karats.
"""
delay(127000)
drawBox(10,20,60,80,72);
drawBox(Var[10],Var[20],Var[60],Var[80],Var[72])
drawObject(16,setXY(12,14))
drawObject(16)
loadRoomWithEgo(43,7,696,49)
pickupObject(1058,97)
pickupObject(1058)
setVarRange(Var[178],9,[0,0,0,0,0,0,0,0,0]);
setVarRange(Var[178],[0,0,0,0,0,0,0,0,0]);
setVarRange(Var[178],9,[256,0,257,0,258,0,259,0,260]);
restoreVerbs(200,209,1)
systemOps(3)
PseudoRoom(98,12,16,18,23)
""", # Test 19: quirky functions
"""
Script# 201
[0000] (2C) CursorHide()
[0002] (2C) UserputOff()
[0004] (1A) VAR_VERB_SCRIPT = 14;
[0009] (0A) startScript(16,[5])
""", # Test 20: defining preprocessor info
"""
Script# 201
[0000] (A8) if (Bit[321]) {
[0005] (11)   animateCostume(8,250)
[0008] (13)   ActorOps(8,[Init(),Costume(21),IgnoreBoxes(),SetZClip(1)]);
[0011] (2D)   putActorInRoom(8,2)
[0014] (01)   putActor(8,198,30)
[001A] (5D)   setClass(20,[32])
[0021] (18) } else {
[0024] (2D)   putActorInRoom(8,0)
[0027] (5D)   setClass(20,[160])
[002E] (**) }
[002E] (A0) stopObjectCode()
END
""", # Test 21: This is an entire (small) script!
"""
[058D] (28)     if (!Bit[5 + Var[1]]) {
[059B] (19)         doSentence(STOP)
[05AD] (**)     }
""", # Test 22: Indirect Bit variable
""" 
[0000] (48) if (Local[1] == 1) {
[0007] (14)   print(255,[Text(" ")]);
[000C] (60)   freezeScripts(127)
[000E] (33)   saveLoad?(1,26))
[0012] (80)   breakHere()
[0013] (48)   if (VAR_GAME_LOADED == 203) {
[001A] (1A)     VAR_MAINMENU_KEY = 319;
[001F] (1A)     Var[107] = 1;
[0024] (60)     freezeScripts(0)
[0026] (62)     stopScript(0)
[0028] (18)   } else {
[002B] (1A)     VAR_MAINMENU_KEY = 0;
[0030] (1A)     Var[107] = 0;
[0035] (F2)     loadRoom(Local[0])
[0038] (**)   }
[0038] (18) } else {
[003B] (2C)   CursorHide()
[003D] (33)   saveLoad?(2,26))
[0041] (80)   breakHere()
[0042] (**) }
[0042] (A0) stopObjectCode()
""", # Test 23: problematic section
"""
for ( Var[109] = 0 ; Var[109] <= 40 ; Var[109] += 1 ) {
 CursorHide()
}
"""
]
    
    knownfailures = frozenset(["07"]) # These tests are supposed to fail
    
    failedtests = [] # keep track of test that have failed
    
    for i, t in enumerate(testlist):
        #if i != 23: # used for debugging specific problems
            #continue
        
        istr = str(i + 1).zfill(2)
        jumpTable = {}
        labelTable = {}
        branchStack = []
        lineNum = 0
        branchCounter = 0
        
        for k, v in actionMap.items():
            k.setParseAction(v)
        
        try:
            print "Test " + istr + ":\n>> " + t.replace("\n", "\n   ")
            results = grMainBlockTesting.parseString(t)
            
            resolve_jumps(results)
            
            if not istr in knownfailures:
                print "== Results: " + str(results) + "\n"
            else:
                print "!! ERRONEOUSLY SUCCEEDED: " + str(results) + "\n"
        except ParseException, pe:
            if not istr in knownfailures:
                print "!! FAILED: " + str(pe) + "... lineNum: " + str(lineNum) + "\n" + pe.line
                failedtests.append(istr)
            else:
                print "== Failed: " + str(pe) + "... lineNum: " + str(lineNum) + "\n" 
    
    if len(failedtests) > 0:
        print "\nThe following tests failed: " + ', '.join(failedtests)


def do_main():
    global jumpTable
    global labelTable
    global lineNum
    global actionMap
    global scriptType
    
    returnval = 0

    if len(sys.argv) == 1:
        print "Please give at least one file name as the argument to Scummbler."
        return 1
    
    for infile in sys.argv:
        if not os.path.isfile(infile):
            print "Invalid filename: " + str(infile)
            returnval = 1
            continue
        
        try:
            jumpTable = {}
            labelTable = {}
            lineNum = 0
        
            for k, v in actionMap.items():
                k.setParseAction(v)
            
            try:
                print "Parsing \"" + infile + "\"..."
                results = grMainBlock.parseFile(infile)
                print "Resolving jumps..."
                resolve_jumps(results)
                
            except ParseException, pe:
                print "Error parsing input file: " + str(pe)
                print pe.line
                returnval = 1
                continue
            except ScummblerException, se:
                print "Error parsing input file: " + str(se)
                returnval = 1
                continue
            
            header = scriptType
        
            outtext = ''.join(results)
            textlen = len(outtext) + 8
            # Add length of the file to the header (Big Endian)
            header += (chr((textlen & 0xFF000000) >> 24) + 
                       chr((textlen & 0xFF0000) >> 16) +
                       chr((textlen & 0xFF00) >> 8) +
                       chr((textlen & 0xFF)))
            
            print "Writing to " + os.path.splitext(infile)[0] + ".out." + scriptType
            outfile = file(os.path.splitext(infile)[0] + ".out." + scriptType, "wb")
            outfile.write(header + outtext)
            outfile.close()
            print "Done!"
        except Exception, e:
            returnval = 1
            print "ERROR - Unhandled Exception: " + str(e)
        
    return returnval
    

if __name__ == "__main__":
    do_main()
    #do_test()
