#!/usr/bin/env python
"""
Scummbler v2

    This file is part of Scummbler.

    Scummbler is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Scummbler is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Scummbler.  If not, see <http://www.gnu.org/licenses/>.

Laurence Dougal Myers
www.jestarjokin.net

December 2008 - May 2009

Atomic parsing elements
"""
import string
from pyparsing import *
import scummbler_vars
from scummbler_misc import definedValues, to_byte, ScummblerParseException # required for the parse actions here

# le = lexicon entry, also gives the code a French feel

# Numbers
leUnsignedInt = Word(nums)
leInt = Word(nums + "-", nums)
leHexByte = Word(nums + srange("[a-fA-F]"), exact=2) # for descummed scripts
leHexWord = Word(nums + srange("[a-fA-F]"), exact=4) # for descummed scripts
leDefinedValueDef = Word(alphanums + "_") # prevents parsing on definition
leDefinedValue = Word("$", alphanums + "_") # compiler define value (I hate using the $)
leByte = (leInt | leDefinedValue)("value")
leWord = (leInt | leDefinedValue)("value")
leDelay = leUnsignedInt
leLabel = Word(alphanums + "_")
leLocalIndex = leUnsignedInt
leBitIndex = leUnsignedInt # same for now
leGlobalIndex = leUnsignedInt # same for now
leLocalIndexOrPointer = Forward() # I'm not sure if this can be indirect
leBitIndexOrPointer = Forward() # I'm not sure if this can be indirect
#leLocalIndexOrPointer = leLocalIndex
#leBitIndexOrPointer = leBitIndex
leGlobalIndexOrPointer = Forward()


# Variables (these could probably be condensed)
leLocalVar = Literal("Local")("vartype") + Suppress(Literal("[")) + leLocalIndexOrPointer("value") + Suppress(Literal("]"))
leBitVar = Literal("Bit")("vartype") + Suppress(Literal("[")) + leBitIndexOrPointer("value") + Suppress(Literal("]"))
leGlobalVar = Literal("Var")("vartype") + Suppress(Literal("[")) + leGlobalIndexOrPointer("value") + Suppress(Literal("]"))
leKnownGlobalVar = oneOf(" ".join( (vn for vn in scummbler_vars.varNames5list if vn != None) ))("knownvar")
# TODO: add support for indirect pointers with known variables
#leKnownGlobalVar = oneOf(" ".join( (vn for vn in scummbler_vars.varNames5list if vn != None) ))("knownvar") + Literal("+") + Group(leVarOrWord)("offset")

leVariable = leLocalVar | leBitVar | leGlobalVar | leKnownGlobalVar | leDefinedValue
leVarOrByte = leVariable | leByte
leVarOrWord = leVariable | leWord
leListItem = Group(leVariable | leWord) # group is a hack to work around nested vars etc

leIndexPointer = Group(leUnsignedInt("base") + Literal("+") + Group(leVarOrWord)("offset"))

# Numbers (Forwarded/Recursive)
leLocalIndexOrPointer << (leIndexPointer | leLocalIndex) # removed because only globals can be indirect
leBitIndexOrPointer << (leIndexPointer | leBitIndex) # removed because only globals can be indirect
leGlobalIndexOrPointer << (leIndexPointer | leGlobalIndex)

leArgList = Suppress(Literal("[")) + Optional(delimitedList(leListItem))("list") + Suppress(Literal("]"))

# Other
#leString = QuotedString(quoteChar="\"", escChar="^")
leString = quotedString # this is the only atom that has its own parse action, due to descumm's poor handling of escape chars
#leString00 = leString
#leStringFF = leString
leComment = cStyleComment # luckily, descumm uses C-style comments!

# descumm uses the karat char ("^") to escape characters.
# Unfortunately, SCUMM scripts use the "^" char to represent three ellipses...
#  and descumm does not escape it!
# So, I've tried to work around it with this parse action.
# NOTE: this means strings will NOT WORK if they have an escape character followed
#  by a digit! We can make it work for 3-digit escape codes, but any less and
#  we don't know where the escape code ends and the digit in the text begins.
#  Actually, descumm only outputs escape codes if (i <= 31 && i >= 128), so
#  we could check if:
#  code is > 1000 (one too many digits)
#  code is > 31 and < 128 (not output by descumm, so assume it's invalid)
def do_leString(s, loc, toks):
    esc = False
    esc_val = ""
    out_string = ""
    for c in toks[0]:
        if esc: # we're currently escaping
            if c in string.digits: # digit following ^
                esc_val += c
                continue
            else: # not a digit following ^, either stop escaping or print "^"
                esc = False
                if esc_val == "": # no digits following "^", assume it's the karat char
                    out_string += "^"
                elif int(esc_val) > 255 or int(esc_val) < 0: #or \
                    #(int(esc_val) > 31 and int(esc_val) < 128): # check if escape code is valid
                    raise Exception("Invalid escape code: ^" + esc_val)
                else: # add escape code to output string
                    out_string += chr(int(esc_val))
        
        if c == "^": # start escaping
            esc = True
            esc_val = ""
        else: # not escaping at all, just add the char
            out_string += c
                
    return out_string[1:-1] # trim surrounding quote marks
leString.setParseAction(do_leString)


def do_leDefinedValue(s, loc, toks):
    global definedValues # very dodgy, defined in scummbler.py, so this will fail to run by itself
    name = toks[0].lstrip('$')
    if not name in definedValues:
        raise ScummblerParseException("Unknown variable name: " + name)
    #return leVarOrWord.parseString(definedValues[name])
    return definedValues[name]
leDefinedValue.setParseAction(do_leDefinedValue)
